#include <stdio.h>
#include <sql.h>
#include <sqlext.h>
#include <sqltypes.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/mman.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#define CHUNKSIZE 1024 
#define DEBUG 0

/*
 * see Retrieving ODBC Diagnostics
 * for a definition of extract_error().
 */
static void extract_error( char *fn, SQLHANDLE handle, SQLSMALLINT type) {
    SQLINTEGER	 i = 0;
    SQLINTEGER	 native;
    SQLCHAR	 state[ 7 ];
    SQLCHAR	 text[256];
    SQLSMALLINT	 len;
    SQLRETURN	 ret;

    fprintf(stderr,
            "\n"
            "The driver reported the following diagnostics whilst running "
            "%s\n\n",
            fn);

    do
    {
        ret = SQLGetDiagRec(type, handle, ++i, state, &native, text,
                            sizeof(text), &len );
        if (SQL_SUCCEEDED(ret))
            printf("%s:%ld:%ld:%s\n", state, i, native, text);
    }
    while( ret == SQL_SUCCESS );
}

void handle_sigbus(int err, siginfo_t *info, void *unused) {
	printf("Got a SIGBUS!!!\n");
	if (info->si_code == BUS_ADRALN) {
		printf("Address alignment error\n");
	} else if (info->si_code == BUS_ADRERR) {
		printf("Address out of bounds\n");
	} else if (info->si_code == BUS_OBJERR) {
		printf("Machine-specific object error\n");
	} else {
		printf("Unknown cause: %d\n", info->si_code);
	}
	exit(1);
}

struct sigaction sigbus = {
	.sa_sigaction = handle_sigbus,
	.sa_flags = SA_SIGINFO,
};

int main() {
  static SQLHENV env;
  static SQLHDBC dbc;
  static SQLHSTMT stmt;
  static SQLRETURN ret; /* ODBC API return status */
  static SQLCHAR outstr[1024];
  static SQLSMALLINT outstrlen;
  int myres = 0;

  /* Allocate an environment handle */
  SQLAllocHandle(SQL_HANDLE_ENV, SQL_NULL_HANDLE, &env);
  /* We want ODBC 3 support */
  SQLSetEnvAttr(env, SQL_ATTR_ODBC_VERSION, (void *) SQL_OV_ODBC3, 0);
  /* Allocate a connection handle */
  SQLAllocHandle(SQL_HANDLE_DBC, env, &dbc);
  /* Connect to the DSN mydsn */
  ret = SQLDriverConnect(dbc, (void *)1, (unsigned char *)"DSN=testing;", SQL_NTS,
			 outstr, sizeof(outstr), &outstrlen,
			 SQL_DRIVER_COMPLETE);
  if (!SQL_SUCCEEDED(ret)) {
    fprintf(stderr, "Failed to connect\n");
    extract_error("SQLDriverConnect", dbc, SQL_HANDLE_DBC);
  } else {
    SQLSMALLINT columns; /* number of columns in result-set */
    int fd=-1;

    printf("Connected\n");

    /* OK, we've got the information we need... now let's get some more */
    myres = SQLAllocHandle(SQL_HANDLE_STMT, dbc, &stmt);
    if ((myres == SQL_SUCCESS) || (myres== SQL_SUCCESS_WITH_INFO))
    {
      char sql[256] = "SELECT data FROM lo_test";
      int res = 0, i, x;
      SQLINTEGER nativeerror=0, numfields=0;
      SQLSMALLINT diagbytes=0;
      unsigned char state[10], diagnostic[256];
      char coltitle[256];
      SQLSMALLINT collen;
      SQLSMALLINT datatype;
      SQLSMALLINT decimaldigits;
      SQLSMALLINT nullable;
      SQLULEN colsize;
      SQLLEN colsize2;
      /* Yes, this is a big hack -- but it's a useful test */
      char *buffer;

      res = SQLPrepare(stmt, (unsigned char *)sql, SQL_NTS);
      if ((res != SQL_SUCCESS) && (res != SQL_SUCCESS_WITH_INFO)) {
        printf("SQL Prepare failed![%s]\n", sql);
        SQLFreeHandle (SQL_HANDLE_STMT, stmt);
	return(-1);
      } 

      res = SQLExecute(stmt);
      if ((res != SQL_SUCCESS) && (res != SQL_SUCCESS_WITH_INFO) && (res != SQL_NO_DATA)) {
        if (res == SQL_ERROR) {
          SQLGetDiagField(SQL_HANDLE_STMT, stmt, 1, SQL_DIAG_NUMBER, &numfields, SQL_IS_INTEGER, &diagbytes);
          for (i=0; i< numfields + 1; i++) {
            SQLGetDiagRec(SQL_HANDLE_STMT, stmt, i + 1, state, &nativeerror, diagnostic, sizeof(diagnostic), &diagbytes);
            printf("SQL Execute returned an error %d: %s: %s (%d)\n", res, state, diagnostic, diagbytes);
            if (i > 10) {
              printf("Oh, that was good.  There are really %d diagnostics?\n", (int)numfields);
              break;
            }
          }
        }
      } 
      if (res == SQL_SUCCESS_WITH_INFO) {
        extract_error("SQLDriverConnect", dbc, SQL_HANDLE_DBC);
      }

      /* OK, we're still good to go! */
      res = SQLFetch(stmt);
      if (res == SQL_NO_DATA) {
        printf("We got no data\n");
      }
      fd = open("/tmp/george.gsm", O_RDWR | O_CREAT | O_TRUNC, 0770);
      if (fd < 0) {
        printf("We failed to open the file for writing!\n");
	return(-2);
      }
      res = SQLNumResultCols(stmt, &columns);
      printf("columns = %d\n",columns);

      for (x=0;x<columns;x++) {
        collen = sizeof(coltitle);
        res = SQLDescribeCol(stmt, x + 1, (unsigned char *)coltitle, sizeof(coltitle), &collen, 
						&datatype, &colsize, &decimaldigits, &nullable);
        printf("res = %d\n",res);
        printf("coltitle = %s\n",coltitle);
        printf("collen = %d\n",collen);
        printf("colsize = %d\n",(int)colsize);
        printf("datatype = %d\n",datatype);
      }
      /* We should really make all the following part of the for loop,
         but since we're only pulling one column, it really doesn't 
         matter.  Just watch out for the "x" below -- it needs to be 
         "x+1" inside of the for loop */

      if (!strcasecmp(coltitle, "data")) {
#if DEBUG
        /* Pass in a NULL buffer, so we can get the size of what we'd like to pull out */
        res = SQLGetData(stmt, x, SQL_BINARY, NULL, 0, &colsize2);
		printf("The object is %d bytes!\n",colsize2);
#endif
		sigaction(SIGBUS, &sigbus, NULL);

		buffer = mmap(NULL, 1025, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
		if (buffer == (char *)-1) {
			printf("mmap failed!\n");
			exit(1);
    	}

        res = SQLGetData(stmt, x, SQL_BINARY, buffer, 1024, &colsize2);
		printf("The object is %d bytes!\n",(int)colsize2);

		munmap(buffer, 1025);
        close(fd);
        truncate("/tmp/george.gsm", colsize2);
        printf("All done!\n");

      }
      SQLFreeHandle(SQL_HANDLE_STMT, stmt);
    }
    SQLDisconnect(dbc);		/* disconnect from driver */
  }
  /* free up allocated handles */
  SQLFreeHandle(SQL_HANDLE_DBC, dbc);
  SQLFreeHandle(SQL_HANDLE_ENV, env);

  return(0);
}
